(* Correction du TP 6 : Automates finis *)

(************)
(*** Oups ***)
(************)

type automate = { transition: (char * int) list vect; final: bool vect};;

let robot = { transition=[| [`Y`, 1; `M`, 2]; [`O`, 3]; [`A`, 3]; [`R`, 4]; [`E`, 5]; [`L`, 6]; [] |]; final=[|false; false; false; false; false; false; true|] };;


(*************************************)
(*** Automates finis déterministes ***)
(*************************************)

(* Question 1 *)
(* L'automate en question est :
 *       +---+
 *       | 1 |
 *       +---+
 *        | ^
 *       a| |a
 *        v |
 *       +---+
 *   --->| 0 |--->
 *       +---+
 *        ^ |
 *       b| |b
 *        | v
 *       +---+
 *       | 2 |
 *       +---+
 *)
let double = { transition=[| [`a`, 1; `b`, 2]; [`a`, 0]; [`b`, 0] |]; final=[|true; false; false|] };;

(* Question 2 *)
let accepte dfa mot =
	let n=string_length mot
	in let rec aux pos ind =
		if ind=n
			then dfa.final.(pos)
			else try aux (assoc mot.[ind] dfa.transition.(pos)) (ind - 1)
				with Not_found -> false
	in aux 0 0;;
(* Notez que cette fonction n'est pas récursive terminale à cause du bloc try. *)

(* Question 3 *)
(* Il s'agit d'un simple parcours de graphe, un parcours en profondeur ici. *)
let langage_vide dfa =
	let n = vect_length dfa.transition
	in let visit = make_vect n false
	in let rec explore l = match l with
		| [] -> false
		| (_,h)::t -> if dfa.final.(h)
			then true
			else ((if visit.(h)
					then false
					else begin
						visit.(h) <- true;
						explore dfa.transition.(h)
					end)
				|| explore t)
	in not explore [(`0`, 0)];;

(* Question 4 *)
(* Un autre parcours de graphe, nécessairement un parcours en largeur cette fois ci. *)
#open "queue";;
let minimum_acceptable dfa pref =
	let rec exec mot pos =
		let n=string_length mot
		in if n=0
			then Some pos
			else try exec (sub_string mot 1 (n-1)) (assoc mot.[0] dfa.transition.(pos))
				with Not_found -> None
	in let bfs pos =
		let n = vect_length dfa.transition
		and queue = new ()
		and ret = ref None
		in let visit = make_vect n false
		in begin
			add (pos, pref) queue;
			while length queue > 0 do
				let (cur,mot) = take queue
				in if dfa.final.(cur)
					then begin
						ret := Some mot;
						clear queue
					end else do_list (function (l,c) -> if not visit.(c)
					then begin
						visit.(c) <- true;
						add (c, mot^ string_of_char l) queue
					end) dfa.transition.(cur)
			done;
			!ret
		end
	in match exec pref 0 with
		| None -> None
		| Some p -> bfs p;;
#close "queue";;


(*****************************************)
(*** Automates finis non déterministes ***)
(*****************************************)

(* Question 5 *)
let poly_assoc elem l =
	let rec aux l acc = match l with
		| [] -> acc
		| (a,b)::q -> aux q (if a=elem then b::acc else acc)
	in aux l [];;

(* Question 6 *)
let naccepte nfa mot =
	let n = string_length mot
	in let rec aux etats ind =
		if n=ind
			then exists (function s -> nfa.final.(s)) etats
			else aux (flat_map (function s -> poly_assoc mot.[ind] nfa.transition.(s)) etats) (ind - 1)
	in aux [0] mot;;

(* Question 7 *)
let intersection nfa1 nfa2 =
	let n1 = vect_length nfa1.transition
	and n2 = vect_length nfa2.transition
	in let n = n1*n2
	in let result = { transition = make_vect n []; final = make_vect n false; }
	and index i1 i2 = i1*n2 + i2
	in begin
		for i1=0 to n1-1 do
			for i2=0 to n2-1 do
			 result.final.(index i1 i2) <- nfa1.final.(i1) && nfa2.final.(i2);

			 let rec childs l1 acc = match l1 with 
				| [] -> acc
				| (l, s1)::q -> childs q (map (function s -> (l, index s1 s)) (poly_assoc l nfa2.transition.(i2)) @ acc)
			 in result.transition.(index i1 i2) <- childs nfa1.transition.(i1) []
			done
		done;
		result
	end;;

(* Question 8 *)
(* Ici, on va s'assurer que tout les états de l'automate sont
 * atteignable.
 * Dans le pire cas, l'automate determinisé aura 2^n états, il est donc
 * impossible d'obtenir une meilleure complexité que du O(2^n), 
 * 
 * La fonction child_list permet de récupérer une liste dont les
 * éléments sont de la forme (lettre, liste d'états).
 * La fonction construct remplit l'automate "raw" de taille 2^n, c'est
 * un parcours en profondeur.
 * La fonction build construit l'automate "res" qui est une version
 * réduite de l'automate "raw".
 *)
let exp a n = 
	let rec aux p a n = match n with
		| 0 -> p
		| n when n mod 2 == 0 -> aux p (a * a) (n / 2)
		| n -> aux (p * a) (a * a) (n / 2)
	in aux 1 a n;;

#open "sort";;
let determinise nfa =
	let child_list l =
		let rec aux l cur tmp acc = match l with
			| [] -> ((cur,tmp)::acc)
			| (hl, hs)::q -> if hl=cur
				then aux q cur (if hs<>hd tmp then hs::tmp else tmp) acc
				else aux q hl [hs] ((cur,tmp)::acc)
		in let sort_predicate (la, sa) (lb, sb) = if int_of_char la = int_of_char lb then sa < sb else int_of_char la < int_of_char lb
		in let sl = sort sort_predicate l
		in aux (tl sl) (fst (hd sl)) [(snd (hd sl))] []
	in let tn=exp 2 (vect_length nfa.transition)
	in let raw = make_vect tn None
	and index l = it_list (fun a b -> a + exp 2 b) 0 l
	and n = ref 0
	and inds = ref []
	in let rec construct l =
		let cl = child_list (flat_map (function s -> nfa.transition.(s)) l)
		in begin
			raw.(index l) <- Some (!n, l, cl);
			incr n;
			inds:=(index l)::!inds;
			do_list (function (l,sl) -> if raw.(index sl) = None then (raw.(index sl) <- Some (0,[],[]); construct sl) else ()) cl
		end
	in begin
		construct [0];
		let res = { transition=make_vect !n []; final=make_vect !n false }
		in let rec build l i = match l with
			| [] -> res
			| h::q -> match raw.(h) with
				| None -> failwith "Impossible"
				| Some (_, tl, cl) -> begin
					let id_of_sl (l, sl) = match raw.(index sl) with
						| None -> failwith "Impossible"
						| Some (id, _, _) -> (l, id)
					in res.transition.(i) <- map id_of_sl cl;
					res.final.(i) <- exists (function s -> nfa.final.(s)) tl;
					build q (i+1)
				end
		in build (rev !inds) 0
	end;;
#close "sort";;
